@@ -1,12 +1,14 @@
 import io
+import itertools
 try:
     from collections import OrderedDict
 except ImportError:
     from ordereddict import OrderedDict
 
-from .xmlcommon import parse_xml, XmlNode
+from .xmlcommon import parse_xml, XmlNode, parse_xml_lazy
 from . import dsd
-    
+from .iteration import EagerIteration, LazyIteration
+
 
 class DsdFetcher(object):
     def __init__(self, requests):
@@ -33,47 +35,58 @@ class Observation(object):
         self.value = value
 
 
-def data_message_reader(parser, fileobj, requests=None, dsd_fileobj=None):
-    class DatasetsReader(object):
-        def __init__(self, tree, dsd_fetcher):
-            self._tree = tree
+def data_message_reader(parser, fileobj, lazy=None, requests=None, dsd_fileobj=None):
+    if lazy:
+        iteration = LazyIteration
+    else:
+        iteration = EagerIteration
+    
+    if dsd_fileobj is None:
+        default_dsd_reader = None
+    else:
+        default_dsd_reader = dsd.reader(fileobj=dsd_fileobj)
+    
+    class MessageReader(object):
+        def __init__(self, root, dsd_fetcher):
+            self._root = root
             self._dsd_fetcher = dsd_fetcher
         
         def datasets(self):
-            return map(
+            return iteration.map(
                 self._read_dataset_element,
-                parser.get_dataset_elements(self._tree),
+                parser.get_dataset_elements(self._root),
             )
             
         def _read_dataset_element(self, element):
-            return DatasetReader(element, self._dsd_fetcher)
-
-    class DatasetReader(object):
-        def __init__(self, element, dsd_fetcher):
-            self._element = element
-            self._dsd_fetcher = dsd_fetcher
+            return DatasetReader(element, self._key_family(element))
         
-        def key_family(self):
-            dsd_reader = self._dsd_reader()
+        def _key_family(self, dataset_element):
+            dsd_reader = self._dsd_reader(dataset_element)
             return KeyFamily(
-                parser.key_family_for_dataset(self._element, dsd_reader),
+                parser.key_family_for_dataset(dataset_element, dsd_reader),
                 dsd_reader,
             )
-        
-        def series(self):
-            key_family = self.key_family()
-            return map(
-                lambda args: self._read_series_element(key_family, *args),
-                parser.get_series_elements(self._element),
-            )
             
-        def _dsd_reader(self):
-            key_family_uri = self._element.get("keyFamilyURI")
+        def _dsd_reader(self, dataset_element):
+            key_family_uri = dataset_element.get("keyFamilyURI")
             if key_family_uri is None:
-                dsd_fileobj.seek(0)
-                return dsd.reader(fileobj=dsd_fileobj)
+                return default_dsd_reader
             else:
                 return self._dsd_fetcher.fetch(key_family_uri)
+
+    class DatasetReader(object):
+        def __init__(self, element, key_family):
+            self._element = element
+            self._key_family = key_family
+        
+        def key_family(self):
+            return self._key_family
+        
+        def series(self):
+            return iteration.map(
+                lambda args: self._read_series_element(self._key_family, *args),
+                parser.get_series_elements(self._element),
+            )
         
         def _read_series_element(self, key_family, element, key):
             return SeriesReader(key_family, element, key)
@@ -153,13 +166,17 @@ def data_message_reader(parser, fileobj, requests=None, dsd_fileobj=None):
                         code, = codes
                         return code
                 
-                return [
+                return (
                     Observation(time=describe_time_code(observation.time), value=observation.value)
                     for observation in observations
-                ]
+                )
             else:
                 return observations
 
-    tree = XmlNode(parse_xml(fileobj).getroot())
+    if lazy:
+        root = parse_xml_lazy(fileobj)
+    else:
+        root = XmlNode(parse_xml(fileobj).getroot())
+    
     dsd_fetcher = DsdFetcher(requests)
-    return DatasetsReader(tree, dsd_fetcher=dsd_fetcher)
+    return MessageReader(root, dsd_fetcher=dsd_fetcher)
